/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Cycle Collector.
 *
 * The Initial Developer of the Original Code is
 * Mozilla Foundation.
 * Portions created by the Initial Developer are Copyright (C) 2010
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "nsCycleCollectorUtils.h"

#include "prthread.h"
#include "nsCOMPtr.h"
#include "nsServiceManagerUtils.h"
#include "nsXPCOMCIDInternal.h"

#include "nsIThreadManager.h"

#if defined(XP_WIN)
#include "windows.h"
#endif

#ifndef MOZILLA_INTERNAL_API

bool
NS_IsCycleCollectorThread()
{
  bool result = false;
  nsCOMPtr<nsIThreadManager> mgr =
    do_GetService(NS_THREADMANAGER_CONTRACTID);
  if (mgr)
    mgr->GetIsCycleCollectorThread(&result);
  return bool(result);
}

#elif defined(XP_WIN)

// Defined in nsThreadManager.cpp.
extern DWORD gTLSThreadIDIndex;

bool
NS_IsCycleCollectorThread()
{
  return TlsGetValue(gTLSThreadIDIndex) ==
    (void*) mozilla::threads::CycleCollector;
}

#elif !defined(NS_TLS)

// Defined in nsCycleCollector.cpp
extern PRThread* gCycleCollectorThread;

bool
NS_IsCycleCollectorThread()
{
  return PR_GetCurrentThread() == gCycleCollectorThread;
}

#endif

#ifdef WINE_GECKO_SRC

#include <new>
#include "nsCycleCollectionParticipant.h"

struct CCObjCallback {
    void (NS_STDCALL *unmark_if_purple)(nsISupports*);
    nsresult (NS_STDCALL *traverse)(nsCycleCollectionParticipant*,void*,nsCycleCollectionTraversalCallback&);
    nsresult (NS_STDCALL *unlink)(void*);
};

class CCParticipant : public nsXPCOMCycleCollectionParticipant {
public:
    CCParticipant(const CCObjCallback *callback) : callback(callback) {}

    NS_IMETHOD Traverse(void *p, nsCycleCollectionTraversalCallback &cb);
    NS_IMETHOD_(void) UnmarkIfPurple(nsISupports *p);
    NS_IMETHOD Unlink(void *p);

private:
    const CCObjCallback *callback;
};

void CCParticipant::UnmarkIfPurple(nsISupports *p)
{
    callback->unmark_if_purple(p);
}

nsresult CCParticipant::Traverse(void *p, nsCycleCollectionTraversalCallback &cb)
{
    return callback->traverse(this, p, cb);
}

nsresult CCParticipant::Unlink(void *p)
{
    return callback->unlink(p);
}

extern "C" {
    nsrefcnt __declspec(dllexport) ccref_incr(nsCycleCollectingAutoRefCnt *ref, nsISupports *base)
    {
        return ref->incr(base);
    }

    nsrefcnt __declspec(dllexport) ccref_decr(nsCycleCollectingAutoRefCnt *ref, nsISupports *base)
    {
        nsrefcnt count = ref->decr(base);
        if(!count)
            ref->stabilizeForDeletion(base);
        return count;
    }

    void __declspec(dllexport) ccref_init(nsCycleCollectingAutoRefCnt *ref, nsrefcnt refcnt)
    {
        static_assert(sizeof(nsCycleCollectingAutoRefCnt) == sizeof(void*), "Unexpected nsCycleCollectingAutoRefCnt size");
        new (ref) nsCycleCollectingAutoRefCnt(refcnt);
    }

    void __declspec(dllexport) ccref_unmark_if_purple(nsCycleCollectingAutoRefCnt *ref)
    {
        if(ref->IsPurple())
            ref->unmarkPurple();
    }

    void __declspec(dllexport) ccp_init(nsXPCOMCycleCollectionParticipant *ccp, const CCObjCallback *callback)
    {
        static_assert(sizeof(CCParticipant) == 3*sizeof(void*), "unexpected CCParticipant size");

        new (ccp) CCParticipant(callback);
    }

    void __declspec(dllexport) describe_cc_node(nsCycleCollectingAutoRefCnt *ref, size_t obj_size, const char *obj_name,
                                                nsCycleCollectionTraversalCallback *cb)
    {
        cb->DescribeRefCountedNode(ref->get(), obj_size, obj_name);
    }

    void __declspec(dllexport) note_cc_edge(nsISupports *child, const char *name, nsCycleCollectionTraversalCallback *cb)
    {
        if(cb->WantDebugInfo())
            cb->NoteNextEdgeName(name);
        cb->NoteXPCOMChild(child);
    }
}

#endif
